from dataclasses import dataclass
from functools import lru_cache
from math import prod
import torch as th
from vidrial.jit.static.types import Shape, Layout, Int
from vidrial.jit.static.util import torch_dtype_to_c
from vidrial.jit.static.util import layout_from_shape_stride
from vidrial.jit.jit import jit, render
from vidrial.jit.timingcache import ConfigTimingCache
from vidrial.jit.binding import make_binding

from vidrial.kernels.sympow.dimensions import problem_dimensions, op_output_shape
from vidrial.kernels.mma_configurator import SMEM_LIMIT, dtype_to_bytes


# ------------------- Source Code -------------------
@dataclass
class SourceCode:
    T: str
    XSlabShape: Shape # [d,[b1,...]]
    p: int
    d_tile: int
    duplicate_correction: bool
    GXSlab: Layout
    GZSlab: Layout
    b_tile: int
    ZFrgShape: Shape
    SZTileLayout: Layout
    @property
    def template(self):
        return """
#include <cuda.h>
#include <cuda_runtime.h>
#include <iostream>
#include <cute/tensor.hpp>
#include <cutlass/cutlass.h>
#include "kernels/sympow/kernel.cuh"
#include "kernels/sympow/sympow_cfg.cuh"
using namespace cute;
using namespace vidrial;

extern "C" int launch(void* __raw_Z, void* __raw_X) {
    using T = {T};
    auto Z = reinterpret_cast<T*>(__raw_Z);
    auto X = reinterpret_cast<T*>(__raw_X);
    using XSlabShape = {XSlabShape};
    using GXSlab = decltype(static_tree_cast<int64_t>({GXSlab}{}));
    using GZSlab = decltype(static_tree_cast<int64_t>({GZSlab}{}));
    constexpr int p = {p};
    constexpr int d_tile = {d_tile};
    constexpr int b_tile = {b_tile};
    using ZFrgShape = {ZFrgShape};
    constexpr bool duplicate_correction = {duplicate_correction};
    using ZSlabShape = decltype(sympow_shape<p, d_tile>(XSlabShape{}));
    using XTileShape = Shape<Int<d_tile>, Int<b_tile>>; 
    using ZTileShape = decltype(tpow_shape<p>(XTileShape{}));
    using SZTileLayout = {SZTileLayout};
    using ZFrgThr = decltype(zipped_divide(Layout<ZTileShape>{}, ZFrgShape{}));
    auto cfg = SympowCfg<T, p, XSlabShape, XTileShape, ZFrgThr, GZSlab, GXSlab, SZTileLayout>{};
    return launch_tiled_sympow_kernel<duplicate_correction>(cfg, Z, X);
}
    """
    def __str__(self):
        return render(self.template, self.__dict__)

# ------------------- Binding -------------------
@dataclass
class BindingCfg:
    d: int
    bs: tuple[int, ...]
    X_shape: tuple[int, ...]
    Z_shape: tuple[int, ...]
    X_stride: tuple[int, ...]
    Z_stride: tuple[int, ...]
    power: int
    d_tile: int
    duplicate_correction: bool
    # Performance parameters (don't affect the output)
    b_tile: int
    ZFrgShape: Shape
    SZTileLayout: Layout
    dtype: th.dtype
    @classmethod
    def from_args(cls, Z, X, power, d_tile, duplicate_correction, b_tile, ZFrgShape, SZTileLayout):
        assert Z.device.type == X.device.type == 'cuda', f"Invalid {Z.device}, {X.device}."
        assert Z.dtype == X.dtype
        dtype = Z.dtype
        bs, d, D = problem_dimensions(Z.shape, X.shape, power, d_tile)
        assert Z.shape == op_output_shape(X.shape, power, d_tile)
        return BindingCfg(d, bs, X.shape, Z.shape, X.stride(), Z.stride(),
                          power, d_tile, duplicate_correction, b_tile, ZFrgShape, SZTileLayout, dtype)
    @property
    def source(self):
        """ Conversion between the different variable name and shape conventions
                D = sympow_dim(d, power, d_tile)
                    Python                     c++
                X_shape=[P,D]      ->      GXSlabShape=[D,P]
                Z_shape=[P,tile_num, d_tile, ...]  ->  GZSlabShape=[[d_tile,...],tile_num],P]
        """
        p = self.power
        XSlabShape = Shape(Int(self.d), Shape(*[Int(b) for b in self.bs]))
        batch_num = len(self.bs)
        batch_dims = tuple(range(batch_num))
        X_py2cuda = (batch_num, batch_dims)
        Z_py2cuda = ((tuple(range(p+batch_num, batch_num, -1)), batch_num), batch_dims)
        GXSlab = layout_from_shape_stride(self.X_shape, self.X_stride, X_py2cuda)
        GZSlab = layout_from_shape_stride(self.Z_shape, self.Z_stride, Z_py2cuda)
        sc = SourceCode(
            T=torch_dtype_to_c(self.dtype),
            XSlabShape=XSlabShape,
            p=self.power,
            d_tile=self.d_tile,
            duplicate_correction=self.duplicate_correction,
            GXSlab=GXSlab,
            GZSlab=GZSlab,
            b_tile=self.b_tile,
            ZFrgShape=self.ZFrgShape,
            SZTileLayout=self.SZTileLayout,
        )
        return sc

def binding(Z, X, power, d_tile, duplicate_correction, b_tile, ZFrgShape, SZTileLayout):
    binding_cfg = BindingCfg.from_args(Z, X, power, d_tile, duplicate_correction, b_tile, ZFrgShape, SZTileLayout)
    jit(name = "sympow",
        code = str(binding_cfg.source),
    )(Z, X)


# ---------------------- Autotune --------------------------------

def reasonable_b_tiles(b):
    return [b_tile for b_tile in [1,4] if b_tile <= b and b % b_tile == 0]

def configurations(args: dict) -> list[dict]:
    Z, X, power, d_tile = args['Z'], args['X'], args['power'], args['d_tile']
    dtype = Z.dtype
    bs,d,D = problem_dimensions(Z.shape, X.shape, power, d_tile)
    configs = []
    for b_tile in reasonable_b_tiles(prod(bs)):
        for num_warp in [1,2,4]:
            SZTileLayout = Layout(Shape(Shape(*[Int(d_tile)]*power), Int(b_tile)))
            for b_frag in [b_frag for b_frag in [1,2,4] if b_tile % b_frag == 0]:
                for d_frag in [d_frag for d_frag in [1,2,4] if d_tile % d_frag == 0]:
                    configs.append({
                        'b_tile': b_tile,
                        'ZFrgShape': Shape(Shape(*[Int(d_frag)]*power), Int(b_frag)),
                        'SZTileLayout': SZTileLayout
                    })
    # Filter out configs that use too much SMEM
    smem_used = [c['SZTileLayout'].shape.size() * dtype_to_bytes(dtype) for c in configs]
    # TODO: if the kernel memory was allocated dynamically we could use 100% of the SMEM instead of 1/2
    configs = [c for c, smem in zip(configs, smem_used) if smem <= SMEM_LIMIT//2]
    return configs

def hash_fn(args: dict) -> str:
    Z, X, power, d_tile = args['Z'], args['X'], args['power'], args['d_tile']
    bs,d,D = problem_dimensions(Z.shape, X.shape, power, d_tile)
    return f"Z.shape={Z.shape}-X.shape={X.shape}-power={power}-d_tile={d_tile}-D={D}-d={d}"

cache = ConfigTimingCache('sympow', hash_fn)
binding_autotuned = make_binding(cache=cache, sweep=configurations)(binding)




import concurrent.futures
import copy
from dataclasses import dataclass
from typing import Optional, Any
from concurrent.futures import ThreadPoolExecutor, as_completed
from vidrial.jit.compiler import compile, compiler_flags, get_kernel_hash, hash_to_hex, get_nvcc_compiler, DEFAULT_INCLUDE_DIRS
from vidrial.jit.runtime import Runtime
from vidrial.jit.package import register_runtime
from vidrial.jit.settings import settings
from tqdm import tqdm
import os
import logging

logger = logging.getLogger(__name__)
DEFAULT_JIT_ROOT = os.path.join(os.path.dirname(os.path.abspath(__file__)), '.jit_cache')

COMPILATION_JOBS = set()

@dataclass(frozen=True)
class CompilationJob:
    code: str
    include_dirs: tuple[str]
    arch: str
    flags: tuple[str]
    path: str

def build_path(name: str, code: str, root: str, flags: list[str] = []):
    signature = f'{name}$${get_kernel_hash(code)}$${code}$${get_nvcc_compiler()}$${flags}'
    name = f'kernel.{name}.{hash_to_hex(signature)}'
    return f'{root}/{name}'


def render(template: str, template_params: dict[str, Any]) -> str:
    """Format a C++ template string with the given template parameters.
    
    Args:
        template: The template string with {key} placeholders
        template_params: Dictionary of template parameter values
        
    Returns:
        The formatted string with all {key} placeholders replaced
    """
    new_template = copy.deepcopy(template)
    for key, value in template_params.items():
        if hasattr(value, 'to_cpp'):
            value = value.to_cpp()
        elif isinstance(value, bool):
            value = 'true' if value else 'false'
        new_template = new_template.replace(f'{{{key}}}', str(value))
    return new_template


def jit(name: str, code: str, include_dirs: Optional[tuple[str]] = None, arch: Optional[str] = None, root: Optional[str] = None, flags: Optional[list[str]] = None) -> Runtime:
    """ Just-in-time compile a function.

    Args:
        name: The name of the function to compile.
        code: The code of the function to compile.
        include_dirs: The include directories to use for the compilation.
        arch: The architecture to use for the compilation.
        root: The root directory to store the compiled function.
        flags: The flags to use for the compilation.

    Returns:
        A Runtime callable that can be used to call the compiled function.
    """
    if flags is None:
        flags = compiler_flags(arch)
    if include_dirs is None:
        include_dirs = DEFAULT_INCLUDE_DIRS
    if root is None:
        root = DEFAULT_JIT_ROOT
    
    path = build_path(name=name, code=code, root=root, flags=flags)
    logger.info(f'Calling jit funciton: path= {path}')
    register_runtime(name, path)
    if settings.precompile:
        logger.debug(f'Adding Compilation Job with path={path}')
        COMPILATION_JOBS.add(CompilationJob(code=code, include_dirs=include_dirs, arch=arch, flags=tuple(flags), path=path))
        return lambda *args, **kwargs: 0 # return dummy runtime with no errors
    if Runtime.is_path_valid(path):
        logger.info(f'Using cached runtime')
        runtime = Runtime(path)
    else:
        logger.info(f'Compiling runtime')
        runtime = compile(code=code, include_dirs=include_dirs, arch=arch, path=path, flags=flags)
    logger.debug(f'Returning runtime')
    return runtime

def precompile():
    if not COMPILATION_JOBS:
        logger.debug("No compilation jobs to precompile")
        return
    num_workers = min(settings.max_workers, len(COMPILATION_JOBS))
    logger.info(f"Precompiling {len(COMPILATION_JOBS)} jobs with {num_workers} workers")
    def precompile_job(job: CompilationJob):
        if not Runtime.is_path_valid(job.path):
            runtime = compile(code=job.code, include_dirs=job.include_dirs, arch=job.arch, path=job.path, flags=job.flags)
    with ThreadPoolExecutor(max_workers=num_workers) as executor:
        futures = [executor.submit(precompile_job, job) for job in COMPILATION_JOBS]
        with tqdm(total=len(futures), desc="Precompiling", disable=not settings.verbose) as pbar:
            for future in as_completed(futures):
                pbar.update(1)
                try:
                    future.result()  # This will raise any exceptions that occurred
                except Exception as e:
                    if not settings.allow_failure:
                        raise e
                    if settings.verbose:
                        logger.warning(f"Error during precompilation: {e}")
                    else:
                        logger.debug(f"Error during precompilation: {e}")
    
    logger.info("Precompilation completed")
    COMPILATION_JOBS.clear()  # Clear the jobs after compilation




import os
import shutil
import sys
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop
from wheel.bdist_wheel import bdist_wheel 


class CustomBdistWheel(bdist_wheel):
    def finalize_options(self):
        super().finalize_options()
        self.root_is_pure = False # platform-specific wheel

    def get_tag(self):
        # Force the tags to be what we want
        return ('py3', 'none', 'manylinux2014_x86_64')
        

class CustomBuildPy(build_py):
    def run(self):
        super().run()
        
        # Only do custom file manipulation when building wheels, not for editable installs
        if not getattr(self, '_called_from_develop', False):
            build_lib = self.build_lib
            vidrial_build_path = os.path.join(build_lib, 'vidrial')
            
            source_release = os.path.join('vidrial', 'release')
            build_release = os.path.join(vidrial_build_path, 'release')
            
            # Build cache directories (temporary, for wheel only)
            build_jit_cache = os.path.join(vidrial_build_path, 'jit', '.jit_cache')
            build_timing_cache = os.path.join(vidrial_build_path, 'jit', '.timing_cache')
            
            if os.path.exists(source_release):
                os.makedirs(build_jit_cache, exist_ok=True)
                os.makedirs(build_timing_cache, exist_ok=True)
                
                # Copy from source release/ to build jit_cache/
                for item in os.listdir(source_release):
                    item_path = os.path.join(source_release, item)
                    if os.path.isdir(item_path):
                        dst_path = os.path.join(build_jit_cache, item)
                        shutil.copytree(item_path, dst_path)
                    elif item.endswith('.yaml'):
                        dst_path = os.path.join(build_timing_cache, item)
                        shutil.copy2(item_path, dst_path)
                
                # Only remove build_release if it exists
                if os.path.exists(build_release):
                    shutil.rmtree(build_release)
                print("Organized kernels for wheel packaging")

            # Cleanup tests
            for dir, _, files in os.walk(vidrial_build_path):
                for file in files:
                    if os.path.basename(file) in ("test.py", "conftest.py", "test.cu"):
                        os.remove(os.path.join(dir, file))


class CustomDevelop(develop):
    def run(self):
        # Mark that we're doing an editable install
        if hasattr(self, 'get_finalized_command'):
            build_py_cmd = self.get_finalized_command('build_py')
            setattr(build_py_cmd, '_called_from_develop', True)
        super().run()


setup(
    packages=find_packages(),
    cmdclass={
        'build_py': CustomBuildPy,
        'bdist_wheel': CustomBdistWheel,
        'develop': CustomDevelop,
    },
    zip_safe=False,
)


from torch.utils._pytree import register_pytree_node
from vidrial.jit.static.int import Int
from vidrial.jit.static.tuples import Tuple, flatten, unflatten

class Shape(Tuple):
    """Represents a shape type with nested dimensions. A shape can be constructed
    from a (mixed) list of ints, shapes, or Ints, but once constructed, all elements
    are either Ints or Shapes."""
    name = "Shape"


def shapify(shape: Shape | Int) -> Shape:
    if isinstance(shape, Int):
        return Shape(shape)
    return shape

def static_shape(t) -> Shape:
    if isinstance(t, tuple) or isinstance(t, list) or isinstance(t, Shape):
        return Shape(*[static_shape(i) if not isinstance(i, (Int, int)) else Int(i) if isinstance(i, int) else i for i in t])
    if isinstance(t, int):
        return Shape(Int(t))
    if isinstance(t, Int):
        return Shape(t)
    raise ValueError(f"Invalid shape type: {type(t)}")


register_pytree_node(Shape, lambda t: (list(flatten(t)), t), unflatten)


import pytest

def pytest_addoption(parser):
    parser.addoption(
        "--heavy", action="store_true", default=False, help="run heavy tests"
    )

def pytest_configure(config):
    config.addinivalue_line(
        "markers", "heavy: mark test as heavy to run"
    )

def pytest_collection_modifyitems(config, items):
    if not config.getoption("--heavy"):
        # --heavy given in cli: do not skip heavy tests
        skip_heavy = pytest.mark.skip(reason="need --heavy option to run")
        for item in items:
            if "heavy" in item.keywords:
                item.add_marker(skip_heavy)
    else:
        # --heavy not given in cli: skip heavy tests
        return 


import torch as th
from itertools import product
from vidrial.jit.settings import settings, PickAny
from flash_attn.flash_attn_interface import flash_attn_func
from vidrial.kernels.flash.op import op as flash_vidrial
from vidrial.py_utils.test_utils import diff
import pandas as pd
import logging
logging.basicConfig(level=logging.INFO)

with settings.set(policy=PickAny, max_configs=999999, plot_timings=False, max_workers=127):
    th.manual_seed(42)
    results = []
    batch_sizes = [1]
    head_sizes = [8]
    seq_lens = [512, 1024]
    head_dims = [32, 128]
    for b, h, tq, d in product(batch_sizes, head_sizes, seq_lens, head_dims):
        tk = tq
        e = d
        Q = th.randn(b, tq, h, d, device='cuda', dtype=th.bfloat16)
        K = th.randn(b, tk, h, d, device='cuda', dtype=th.bfloat16)
        V = th.randn(b, tk, h, e, device='cuda', dtype=th.bfloat16)
        O, l = flash_vidrial(Q, K, V)
        O_flash = flash_attn_func(Q, K, V, causal=True, softmax_scale=None)
        max_diff = th.max(th.abs(O - O_flash))
        results.append({
            'b': b,
            'h': h,
            'tq': tq,
            'tk': tk,
            'd': d,
            'e': e,
            'max_diff': max_diff.item()
        })
    print(pd.DataFrame(results))



import inspect
import copy
from typing import Callable, Any
from functools import wraps
from vidrial.jit.tuner import tune_and_update
from vidrial.jit.timingcache import ConfigTimingCache
from vidrial.jit.package import register_config_cache
from vidrial.jit.settings import settings, PickAny, PickBest
import logging
logger = logging.getLogger(__name__)

def plot_timings(fn: Callable, timings: list[dict]):
    """ Plot the timings of the configurations.
    """
    name = f"timing_{fn.__module__}.{fn.__qualname__}.png"
    try:
        import matplotlib.pyplot as plt
        import matplotlib
        matplotlib.set_loglevel('error')  # Silence all non-error logs
    except Exception as e:
        logger.warning("matplotlib is not installed, skipping timing plot")
        return
    plt.figure(figsize=(10, 6))
    plt.plot(range(len(timings)), [t['time'] for t in timings], label=f'Timing for {fn.__module__}.{fn.__qualname__}')
    plt.xlabel('Configuration Index')
    plt.ylabel('Time (ms)')
    plt.legend()
    plt.savefig(name)

def make_binding(cache: ConfigTimingCache, sweep: list[dict] | Callable) -> Callable:
    """ Decorator for picking the best configuration from a config cache object when running a 
    function, optionally run a sweep of configurations to pick the best one when cache misses. 
    

    Canonical usage:
    ```python
        cache = ConfigTimingCache('fn', lambda args: args['X'].shape)
        def sweep(args: dict) -> list[dict]:
            return [
                {'tile_1': 32, 'tile_2': 32, 'thread_num': 32},
                {'tile_1': 64, 'tile_2': 64, 'thread_num': 64},
            ]

        @make_binding(cache, sweep=sweep)
        def fn(X, Y, tile_1, tile_2, thread_num) -> torch.Tensor:
            ...

        fn(X, Y)
    ```
    where the best configuration is picked from the set of configurations that are provided by the decorator.

    There are 3 places where the the configuration can be picked from:
    1. TuningCache: If a given set of argument hashes to a key that is in the cache, the best configuration is picked from the cache.
    2. Sweeping: If cache misses and a sweep is provided, the decorator will run a sweep of configurations and pick the best one (and optionally update the cache).

    It is also possible to bypass the decorator entirely by calling the function with a complete specification of configuration.

    The above semantics are exemplified below:
    
    ```python
        @make_binding(cache)
        def fn(X: torch.Tensor, Y: torch.Tensor, tile_1: int, tile_2: int, thread_num: int) -> torch.Tensor:
            ...

        fn(X, Y) // rely on TuningCache, fails if no cache is found
    ```

    ```python
        @make_binding(cache, sweep=sweep)
        def fn(X: torch.Tensor, Y: torch.Tensor, tile_1: int, tile_2: int, thread_num: int) -> torch.Tensor:
            ...

        fn(X, Y) // rely on TuningCache first, then sweep if cache misses
    ```

    ```python
        @make_binding(cache)
        def fn(X: torch.Tensor, Y: torch.Tensor, tile_1: int = 32, tile_2: int = 32, thread_num: int = 32) -> torch.Tensor:
            ...

        fn(X, Y) // rely on TuningCache first, then default kwargs if cache misses
    ```

    ```python
        @make_binding(cache, sweep=sweep)
        def fn(X: torch.Tensor, Y: torch.Tensor, tile_1, tile_2, thread_num) -> torch.Tensor:
            ...

        fn(X, Y, tile_1=32, tile_2=32, thread_num=32) // bypass the decorator entirely
    ```

    Args:
        cache: A ConfigTimingCache object to store the cache.
        sweep: A list of configurations or a function that takes the arguments and returns a list of configurations to use if the cache is missed.

    Returns:
        A decorator that wraps the function and returns the best configuration.
    """
    assert cache is not None, "cache must be provided"
    register_config_cache(cache)

    def decorator(fn: Callable) -> Callable:
        log_prefix = "function selector | "
        if hasattr(fn, '__qualname__'): # ensure fn can be a magic mock
            log_prefix += f"{fn.__qualname__} | "
        @wraps(fn)
        def _wrapped(*args, **kwargs) -> Any:
            all_arg_names = list(inspect.signature(fn).parameters.keys())
            _args = dict(zip(all_arg_names, args)) | kwargs

            if len(_args) == len(all_arg_names):
                logger.debug(f"{log_prefix}All the arguments are provided, nothing to configure")
                return fn(**_args)

            if _args in cache: 
                logger.info(f"{log_prefix}TimingCache hit at {cache.location}. Run best config")
                if settings.plot_timings:
                    plot_timings(fn, cache.get_timings(_args))
                best_config = cache[_args]
                return fn(**_args, **best_config) # run the best config

            logger.debug(f"{log_prefix}TimingCache miss at {cache.location}. Generating config sweep")
            configs = sweep(_args) if callable(sweep) else copy.copy(sweep)
            logger.debug(f"{log_prefix}Sweep configs: {configs[:10]}...")

            if isinstance(configs, dict) or len(configs) == 1: # a single config
                config = configs if isinstance(configs, dict) else configs[0]
                logger.debug(f"{log_prefix}Sweep produced a single config")
                return fn(**_args, **config)

            assert isinstance(configs, list)
            logger.debug(f"{log_prefix}Sweep produced {len(configs)} configs")
            if len(configs) > settings.max_configs:
                logger.debug(f"{log_prefix}truncating to {settings.max_configs}")
                configs = configs[0:settings.max_configs]
            logger.debug(f"{log_prefix}Policy: {settings.policy}")
            if settings.policy == PickAny:
                if not settings.allow_failure:
                    logger.debug(f"{log_prefix}PickAny mode: Compilation faliures not allowed. Running first config only.")
                    return fn(**_args, **(configs[0]))
                for i, config in enumerate(configs):
                    logger.debug(f"{log_prefix}PickAny mode: Trying config {i} out of {len(configs)}")
                    try:
                        return fn(**_args, **config)
                    except Exception as e:
                        logger.debug(f"{log_prefix}Config {i} failed with error: {e}")
                        continue
                raise ValueError("No valid configs are provided")

            # run the sweep in parallel across max_workers
            if settings.policy == PickBest:
                logger.debug(f"{log_prefix}PickBest mode: Benchmark sweep in parallel across {settings.max_workers} workers")
                timings = tune_and_update(fn, _args, configs,
                                            cache,
                                            num=10,
                                            allow_failure=settings.allow_failure,
                                            no_side_effect=True,
                                            max_workers=min(settings.max_workers, len(configs)),
                                            verbose=settings.verbose)
                if settings.plot_timings:
                    plot_timings(fn, timings)
                best_config = timings[0]['config']
                return fn(**_args, **best_config)
            raise ValueError("Unable to pick and run a config")
        return _wrapped
    
    return decorator



import pytest
import yaml
import pickle
from functools import reduce
from torch.utils._pytree import tree_map, tree_flatten
from vidrial.jit.static import Shape, Int
from vidrial.jit.static.shape import static_shape

EXAMPLES = [
    (2,),
    (Int(2),),
    (1, 2, 3),
    (1, (2, 3)),
    (1, (2, (3, 4))),
    (1, (2, (3, (4, 5)))),
    (1, (Int(2), (3, (4, 5)))),
    (1, (Int(2), (3, Shape(4, 5)))),
]



@pytest.mark.parametrize("example", EXAMPLES, ids=str)
def test_rank(example):
    t = Shape(*example)
    assert t.rank() == len(example)


@pytest.mark.parametrize("example", EXAMPLES, ids=str)
def test_size(example):
    t = Shape(*example)
    t_flat, _ = tree_flatten(tree_map(lambda x: int(x), t))
    assert t.size() == reduce(lambda x, y: x * y, t_flat, 1)


@pytest.mark.parametrize("example", EXAMPLES, ids=str)
def test_depth(example):
    t = Shape(*example)
    def _depth(tup: tuple | Shape) -> int:
        return max(1 + _depth(x) if isinstance(x, (tuple, Shape)) else 1 for x in tup)
    assert t.depth() == _depth(example)


@pytest.mark.parametrize("example", EXAMPLES, ids=str)
def test_modes(example):
    t = Shape(*example)
    def _modes(tup: tuple | Shape) -> int:
        return sum(_modes(x) if isinstance(x, (tuple, Shape)) else 1 for x in tup)
    assert t.modes() == _modes(example)


@pytest.mark.parametrize("example", EXAMPLES, ids=str)
def test_to_tuple(example):
    t = Shape(*example)
    def _to_tuple(tup: tuple | Shape) -> tuple:
        return tuple(_to_tuple(x) if isinstance(x, (tuple, Shape)) else x for x in tup)
    assert _to_tuple(t) == _to_tuple(example)


@pytest.mark.parametrize("example", EXAMPLES, ids=str)
def test_to_list(example):
    t = Shape(*example)
    def _to_list(tup: tuple | Shape) -> list:
        return [_to_list(x) if isinstance(x, (tuple, Shape)) else x for x in tup]
    assert t.to_list() == _to_list(example)


@pytest.mark.parametrize("example", EXAMPLES, ids=str)
def test_iter(example):
    t = Shape(*example)
    def _check_iter(tup: tuple | Shape, example: tuple) -> bool:
        for x, y in zip(tup, example):
            if isinstance(x, (tuple, Shape)):
                return _check_iter(x, y)
            else:
                assert x == y
        return True
    assert _check_iter(t, example)


@pytest.mark.parametrize("example", EXAMPLES, ids=str)
def test_eq(example):
    t1 = Shape(*example)
    t2 = Shape(*example)
    assert t1 == t2
    assert t1 != Shape(*(example[1:]))
    assert t1 != Shape(*(example[:-1]))
    assert t1 != Shape(*(example[1:] + example[:-1]))
    assert t1 != Shape(*(example[1:] + example[:-1]))


@pytest.mark.parametrize("example", EXAMPLES, ids=str)
def test_pickle(example):
    t = Shape(*example)
    t_pickled = pickle.dumps(t)
    t_unpickled = pickle.loads(t_pickled)
    assert t == t_unpickled


@pytest.mark.parametrize("example", EXAMPLES, ids=str)
def test_yaml(example):
    t = Shape(*example)
    print(yaml.dump(t))
    assert yaml.load(yaml.dump(t), yaml.Loader) == t

EXAMPLES_STATIC_SHAPE = [
    ((2,), Shape(Int(2))),
    (Int(2), Shape(Int(2))),
    ((1, 2, 3), Shape(Int(1), Int(2), Int(3))),
    ((1, (2, 3)), Shape(Int(1), Shape(Int(2), Int(3)))),
    ((1, (2, (3, 4))), Shape(Int(1), Shape(Int(2), Shape(Int(3), Int(4))))),
    ((1, (2, (3, (4, 5)))), Shape(Int(1), Shape(Int(2), Shape(Int(3), Shape(Int(4), Int(5)))))),
    ((1, (Int(2), (3, (4, 5)))), Shape(Int(1), Shape(Int(2), Shape(Int(3), Shape(Int(4), Int(5)))))),
    ((1, (Int(2), (3, Shape(4, 5)))), Shape(Int(1), Shape(Int(2), Shape(Int(3), Shape(Int(4), Int(5)))))),
]


@pytest.mark.parametrize("example", EXAMPLES_STATIC_SHAPE, ids=str)
def test_static_shape(example):
    t = static_shape(example[0])
    assert t == example[1]